/*
 * Copyright (c) 2008 appspy.org, Contributors.
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 
 * Unless required by applicable law or agreed to in writing, software distributed under the 
 * License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, 
 * either express or implied. See the License for the specific language governing permissions 
 * and limitations under the License. */

package org.appspy.admin.service.impl;

import java.sql.Timestamp;
import java.util.Collection;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashSet;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.appspy.admin.jobs.AppspyJobUtils;
import org.appspy.admin.service.SchedulingService;
import org.appspy.admin.service.impl.schedule.params.ParamResolverFactory;
import org.appspy.server.bo.scheduler.JobDeclaration;
import org.appspy.server.bo.scheduler.JobExecution;
import org.appspy.server.bo.scheduler.JobExecutionParam;
import org.appspy.server.bo.scheduler.JobSchedule;
import org.appspy.server.bo.scheduler.JobScheduleParam;
import org.appspy.server.dao.SchedulerDao;
import org.quartz.CronExpression;
import org.quartz.CronTrigger;
import org.quartz.JobDetail;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;
import org.quartz.JobListener;
import org.quartz.Scheduler;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.BeanFactory;
import org.springframework.beans.factory.BeanFactoryAware;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

/**
 * @author Olivier HEDIN / olivier@appspy.org
 */
public class SchedulingServiceImpl implements SchedulingService, InitializingBean, BeanFactoryAware {
	
	protected static final Log sLog = LogFactory.getLog(SchedulingServiceImpl.class);
	
	protected SchedulerDao mSchedulerDao = null;
	protected PlatformTransactionManager mTransactionManager = null;
	protected Scheduler mScheduler = null;
	protected ParamResolverFactory mParamResolverFactory = null;
	protected BeanFactory mBeanFactory = null;
	
	/* (non-Javadoc)
     * @see org.appspy.admin.service.SchedulingService#createJobDescription(org.appspy.server.bo.scheduler.JobDeclaration)
     */
	@Transactional(readOnly=false, propagation=Propagation.REQUIRED)
    public JobDeclaration createJobDeclaration(JobDeclaration jobDeclaration)
    {
		checkJobClass(jobDeclaration);
		
		return mSchedulerDao.createJobDeclaration(jobDeclaration);
    }

	/* (non-Javadoc)
     * @see org.appspy.admin.service.SchedulingService#createJobSchedule(org.appspy.server.bo.scheduler.JobSchedule)
     */
	@Transactional(readOnly=false, propagation=Propagation.REQUIRED)
    public JobSchedule createJobSchedule(JobSchedule jobSchedule)
    {
		checkCronExpression(jobSchedule);
		
		JobSchedule result = mSchedulerDao.createJobSchedule(jobSchedule);

		try
	    {
	    	addJobScheduleToScheduler(result);
	    }
	    catch (Exception ex)
	    {
	    	sLog.error(ex.getMessage(), ex);
	    }
	    
	    return result;
    }

	/* (non-Javadoc)
     * @see org.appspy.admin.service.SchedulingService#deleteJobDeclaration(org.appspy.server.bo.scheduler.JobDeclaration)
     */
	@Transactional(readOnly=false, propagation=Propagation.REQUIRED)
    public void deleteJobDeclaration(JobDeclaration jobDeclaration)
    {
		for (JobSchedule jobSchedule : mSchedulerDao.findAllJobSchedules())
        {
	        if (jobSchedule.getJobDeclaration().getId() == jobDeclaration.getId())
	        {
	        	mSchedulerDao.removeJobSchedule(jobSchedule);
	        }
        }
	    mSchedulerDao.removeJobDeclaration(jobDeclaration);
    }

	/* (non-Javadoc)
     * @see org.appspy.admin.service.SchedulingService#deleteJobExecution(org.appspy.server.bo.scheduler.JobExecution)
     */
	@Transactional(readOnly=false, propagation=Propagation.REQUIRED)
    public void deleteJobExecution(JobExecution jobExecution)
    {
	    mSchedulerDao.removeJobExecution(jobExecution);
    }

	/* (non-Javadoc)
     * @see org.appspy.admin.service.SchedulingService#deleteJobSchedule(org.appspy.server.bo.scheduler.JobSchedule)
     */
	@Transactional(readOnly=false, propagation=Propagation.REQUIRED)
    public void deleteJobSchedule(JobSchedule jobSchedule)
    {
	    mSchedulerDao.removeJobSchedule(jobSchedule);
	    try
	    {
	    	removeJobScheduleFromScheduler(jobSchedule);
	    }
	    catch (Exception ex)
	    {
	    	sLog.error(ex.getMessage(), ex);
	    }
    }

	/* (non-Javadoc)
     * @see org.appspy.admin.service.SchedulingService#getJobDeclarations()
     */
    public Collection<JobDeclaration> getJobDeclarations()
    {
	    return mSchedulerDao.getAllJobDeclarations();
    }

	/* (non-Javadoc)
     * @see org.appspy.admin.service.SchedulingService#getJobExecutions()
     */
    public Collection<JobExecution> getJobExecutions()
    {
	    return mSchedulerDao.findAllJobExecutions();
    }

	/* (non-Javadoc)
     * @see org.appspy.admin.service.SchedulingService#getJobSchedules()
     */
    public Collection<JobSchedule> getJobSchedules()
    {
	    return mSchedulerDao.findAllJobSchedules();
    }

	/* (non-Javadoc)
     * @see org.appspy.admin.service.SchedulingService#updateJobDeclaration(org.appspy.server.bo.scheduler.JobDeclaration)
     */
	@Transactional(readOnly=false, propagation=Propagation.REQUIRED)
    public JobDeclaration updateJobDeclaration(JobDeclaration jobDeclaration)
    {
		checkJobClass(jobDeclaration);
		
	    return mSchedulerDao.updateJobDeclaration(jobDeclaration);
    }

	/**
     * @param jobDeclaration
     */
    protected void checkJobClass(JobDeclaration jobDeclaration)
    {
	    try
		{
			Class.forName(jobDeclaration.getClassName());
		}
		catch (Exception ex)
		{
			throw new RuntimeException(ex.getMessage(), ex);
		}

    }

	/* (non-Javadoc)
     * @see org.appspy.admin.service.SchedulingService#updateJobSchedule(org.appspy.server.bo.scheduler.JobSchedule)
     */
	@Transactional(readOnly=false, propagation=Propagation.REQUIRED)
    public JobSchedule updateJobSchedule(JobSchedule jobSchedule)
    {
		checkCronExpression(jobSchedule);

		JobSchedule result = mSchedulerDao.updateJobSchedule(jobSchedule);
	    
	    try
	    {
	    	updateJobScheduleInScheduler(jobSchedule);
	    }
	    catch (Exception ex)
	    {
	    	sLog.error(ex.getMessage(), ex);
	    }

	    return result;
    }
	
	@Transactional(readOnly=false, propagation=Propagation.REQUIRED)
    public void removeJobExecutionsByDate(final Date date)
    {
    	mSchedulerDao.removeJobExecutionsByDate(date);
    }
    
	/* (non-Javadoc)
     * @see org.springframework.beans.factory.InitializingBean#afterPropertiesSet()
     */
    public void afterPropertiesSet() throws Exception
    {
    	mScheduler.addGlobalJobListener(new JobListener(){

			public String getName()
            {
				return "AppspyJobListener";
            }

			public void jobExecutionVetoed(JobExecutionContext jec)
            {
            }

			public void jobToBeExecuted(JobExecutionContext jec)
            {
				JobExecution jobExecution = new JobExecution();
				String jobName = jec.getJobDetail().getName();
				JobSchedule jobSchedule = mSchedulerDao.findJobScheduleById(Long.valueOf(jobName));
				jobExecution.setJobSchedule(jobSchedule);
				HashSet<JobExecutionParam> jobExecutionParams = new HashSet<JobExecutionParam>();
				for (JobScheduleParam jobScheduleParam : jobSchedule.getParams())
                {
	                JobExecutionParam jobExecutionParam = new JobExecutionParam();
	                jobExecutionParam.setJobExecution(jobExecution);
	                jobExecutionParam.setJobScheduleParam(jobScheduleParam);
	                jobExecutionParam.setValue(mParamResolverFactory.getParamResolver(jobScheduleParam).resolveValue(jobScheduleParam));
	                jobExecutionParams.add(jobExecutionParam);
                }
				jobExecution.setParams(jobExecutionParams);
				jobExecution.setStartTime(new Timestamp(GregorianCalendar.getInstance().getTimeInMillis()));
				mSchedulerDao.createJobExecution(jobExecution);
				
				jec.put(AppspyJobUtils.JOB_EXECUTION, jobExecution);
				jec.put(AppspyJobUtils.BEAN_FACTORY, mBeanFactory);
            }

			public void jobWasExecuted(JobExecutionContext jec, JobExecutionException jex)
            {
				JobExecution jobExecution = (JobExecution) jec.get(AppspyJobUtils.JOB_EXECUTION);
				jobExecution.setEndTime(new Timestamp(GregorianCalendar.getInstance().getTimeInMillis()));
				if (jex != null)
				{
					jobExecution.setExceptionMessage(jex.getMessage());
					sLog.error(jex.getMessage(), jex);
				}
				mSchedulerDao.updateJobExecution(jobExecution);
            }});
    	
    	for (JobSchedule jobSchedule : getJobSchedules())
        {
    		try
    		{
    			addJobScheduleToScheduler(jobSchedule);
    		}
    		catch (Exception ex)
    		{
    			sLog.error(ex.getMessage(), ex);
    		}
        }
    }
    
    protected void addJobScheduleToScheduler(JobSchedule jobSchedule) throws Exception
    {
		String jobName = jobSchedule.getId().toString();
		String groupName = jobSchedule.getJobDeclaration().getId().toString();
		JobDetail jobDetail = new JobDetail(jobName, groupName, Class.forName(jobSchedule.getJobDeclaration().getClassName()), false, false, false);
		
		mScheduler.scheduleJob(jobDetail, new CronTrigger("AppspyTriggerGroup", jobName, jobSchedule.getCronExpression()));
    }
    
    protected void removeJobScheduleFromScheduler(JobSchedule jobSchedule) throws Exception
    {
		String jobName = jobSchedule.getId().toString();
		
		mScheduler.unscheduleJob("AppspyTriggerGroup", jobName);
    }

    protected void updateJobScheduleInScheduler(JobSchedule jobSchedule) throws Exception
    {
		removeJobScheduleFromScheduler(jobSchedule);
		addJobScheduleToScheduler(jobSchedule);
    }

    protected void checkCronExpression(JobSchedule jobSchedule)
    {
	    try
		{
			new CronExpression(jobSchedule.getCronExpression());
		}
		catch (Exception ex)
		{
			throw new RuntimeException(ex.getMessage(), ex);
		}
    }
    
	/**
     * @return the schedulerDao
     */
    public SchedulerDao getSchedulerDao()
    {
    	return mSchedulerDao;
    }

	/**
     * @param schedulerDao the schedulerDao to set
     */
    public void setSchedulerDao(SchedulerDao schedulerDao)
    {
    	mSchedulerDao = schedulerDao;
    }

	/**
	 * @return the transactionManager
	 */
	public PlatformTransactionManager getTransactionManager() {
		return mTransactionManager;
	}

	/**
	 * @param transactionManager the transactionManager to set
	 */
	public void setTransactionManager(PlatformTransactionManager transactionManager) {
		mTransactionManager = transactionManager;
	}

	/**
     * @return the scheduler
     */
    public Scheduler getScheduler()
    {
    	return mScheduler;
    }

	/**
     * @param scheduler the scheduler to set
     */
    public void setScheduler(Scheduler scheduler)
    {
    	mScheduler = scheduler;
    }

	/**
     * @return the paramResolverFactory
     */
    public ParamResolverFactory getParamResolverFactory()
    {
    	return mParamResolverFactory;
    }

	/**
     * @param paramResolverFactory the paramResolverFactory to set
     */
    public void setParamResolverFactory(ParamResolverFactory paramResolverFactory)
    {
    	mParamResolverFactory = paramResolverFactory;
    }

	/* (non-Javadoc)
     * @see org.springframework.beans.factory.BeanFactoryAware#setBeanFactory(org.springframework.beans.factory.BeanFactory)
     */
    public void setBeanFactory(BeanFactory beanFactory) throws BeansException
    {
    	mBeanFactory = beanFactory;
    }
    
    
}
